package serverSide;

import java.io.*;
import java.net.*;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.*;

import javax.net.ssl.SSLServerSocket;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLSocket;

public class ChatServer
{
	private Hashtable<String, Node> clientNames = new Hashtable<String, Node>();
	private Hashtable<String, Node> helperNames = new Hashtable<String, Node>();
	private Hashtable<String, Node> unknownNames = new Hashtable<String, Node>();
	private Hashtable<Node, Node> chats = new Hashtable<Node, Node>();
	private Hashtable<String, Node>	blockList = new Hashtable<String, Node>();
	
	private String configFile = ".\\res\\Config.txt";
	private String authFile = ".\\res\\passwd";
	private PrivateKey privKey;
	private PublicKey pubKey;
	private CipherHelperS ch;
	public int connectionTimeout = 2000; //Disconnect timer for unknown connections to server
	public int maxConcurrentUnknown = 5; //Max Unknown type connections to server
	public int maxConnectionAttempts = 5; //Max failed Authorization attempts by single IP
	public int lockOutTime = 5; //Lock out time in minutes
	public int singleIPConnections = 50;
	
	public ChatServer(int socket) throws Exception
	{
		if (!checkAllFiles())
		{
			System.exit(0);
		}
			
		readConfig();
		pubKey = generateRSA();
		
		ch = new CipherHelperS();
		ch.setPrivKey(privKey);
		
		System.setProperty("javax.net.ssl.keyStore", ".\\res\\mySrvKeystore");
	    System.setProperty("javax.net.ssl.keyStorePassword", "123456");
		SSLServerSocketFactory sslserversocketfactory =
                (SSLServerSocketFactory) SSLServerSocketFactory.getDefault();
        SSLServerSocket welcomeSocket =
                (SSLServerSocket) sslserversocketfactory.createServerSocket(socket);
		//ServerSocket welcomeSocket = new ServerSocket (socket);
		Thread t = new ShutdownHookServer(this, welcomeSocket);
		Runtime.getRuntime().addShutdownHook(t);
		//System.out.println("Prior to while loop");

		while(true) 
		{
			SSLSocket connectionSocket = (SSLSocket) welcomeSocket.accept();

			//Socket connectionSocket = welcomeSocket.accept();
			DataOutputStream outToClient = new DataOutputStream(connectionSocket.getOutputStream());
			BufferedReader inFromClient = new BufferedReader(new InputStreamReader(
					connectionSocket.getInputStream()));
			
			String ipAddress = connectionSocket.getRemoteSocketAddress().toString();
			ipAddress = ipAddress.split(":", 2)[0].toString().substring(1);
			if (!unknownNames.containsKey(ipAddress))
			{
				Node n = new Node();
				n.setName(connectionSocket.getRemoteSocketAddress().toString());
				n.setIpAddress(ipAddress);
				n.setUnknown(1);
				n.getConnectList().add(connectionSocket);
				unknownNames.put(ipAddress, n);
			}
			else if ((unknownNames.containsKey(ipAddress)) && (unknownNames.get(ipAddress).getUnknown() < maxConcurrentUnknown))
			{
				unknownNames.get(ipAddress).setUnknown(unknownNames.get(ipAddress).getUnknown() + 1);
				unknownNames.get(ipAddress).getConnectList().add(connectionSocket);
			}
			else
			{
				//validConnection = false;
				//System.out.println("Too many attempts from: " + ipAddress);
				//connectionSocket.close();
				unknownNames.get(ipAddress).getConnectList().get(0).getInputStream().close();
				unknownNames.get(ipAddress).getConnectList().get(0).close();
				unknownNames.get(ipAddress).getConnectList().remove(0);
				unknownNames.get(ipAddress).setUnknown(unknownNames.get(ipAddress).getUnknown() - 1);
			}
			
			NewThreadIncoming w = new NewThreadIncoming(this, connectionSocket, inFromClient);
			Thread m = new Thread(w);
			w.setThread(m);
			m.start();

		}
	}
	
	public PrivateKey getPrivKey()
	{
		return privKey;
	}

	private PublicKey generateRSA()
	{
		KeyPairGenerator keyPairGenerator;
		KeyPair keyPair;
		try
		{
			keyPairGenerator = KeyPairGenerator.getInstance("RSA");
			keyPairGenerator.initialize(1024);
		    keyPair = keyPairGenerator.genKeyPair();
		    privKey = keyPair.getPrivate();
		    return keyPair.getPublic();
		    
		} catch (NoSuchAlgorithmException e)
		{
			System.out.println("Problems with generating RSA keys.");
			System.out.println("Cannot continue operation without generating keys.");
			System.exit(0);
		}
		return null;
	}

	private boolean checkAllFiles()
	{
		boolean check = true;
		if (!new File(configFile).isFile())
		{
			System.out.println("Config file missing! Supposed to be in .\\res\\Config.txt");
			check = false;
		}
		if (!new File(authFile).isFile())
		{
			System.out.println("Authorization file missing! Supposed to be in .\\res\\passwd");
			check = false;
		}
		if (!new File(".\\res\\mySrvKeystore").isFile())
		{
			System.out.println("keystore not located in .\\res\\keystore");
			check = false;
		}
		return check;
	}

	private void readConfig()
	{
		String line = "";
		
		try 
		{
			FileReader fileReader = new FileReader(configFile);

			BufferedReader bufferedReader = new BufferedReader(fileReader);

			while((line = bufferedReader.readLine()) != null) 
			{
				if (line.startsWith("ConnectionTimeout:"))
				{
					this.connectionTimeout = this.connectionTimeout < Integer.parseInt(line.split(":", 2)[1].trim()) ?
							Integer.parseInt(line.split(":", 2)[1].trim()) : this.connectionTimeout;
				}
				else if (line.startsWith("MaxConcurrentUnknown:"))
					this.maxConcurrentUnknown = Integer.parseInt(line.split(":", 2)[1].trim());
				else if (line.startsWith("AuthenticationAttempts:"))
					this.maxConnectionAttempts = Integer.parseInt(line.split(":", 2)[1].trim());
				else if (line.startsWith("LockoutTime:"))
					this.lockOutTime = Integer.parseInt(line.split(":", 2)[1].trim());
				else if (line.startsWith("SingleIPConnections:"))
					this.singleIPConnections = Integer.parseInt(line.split(":", 2)[1].trim());
			}	

			bufferedReader.close();	
		}
		catch (FileNotFoundException e)
		{
			System.out.println("Config file missing, using default values");
			setDefaultValues();
			
		} catch (Exception e)
		{
			System.out.println("Error reading config file, using default values");
			setDefaultValues();
		}
	}

	private void setDefaultValues()
	{
		this.connectionTimeout = 2000;
		this.maxConcurrentUnknown = 5;
		this.maxConnectionAttempts = 5;
		this.lockOutTime = 5;
		this.singleIPConnections = 50;
	}

	public static void main(String argv[]) throws Exception
	{
		if (argv.length == 1)
		{
			int socket = 0;
			try
			{ //Only purpose is to made a non-static instance to work with
				System.out.println("Server started on port: "+ argv[0]);
				socket = Integer.parseInt(argv[0]);
				new ChatServer(socket);
			}
			catch (NumberFormatException nFE)
			{
				usage();
			}
			catch (SocketException e)
			{}
		}
		else 
		{
			usage();
		}
	}
	
	public void putInTable(String name, DataOutputStream outStream, 
		BufferedReader inStream, int type, PublicKey pubKey)
	{ //Used to keep a up-to-date list of users connected to server
		Node newNode = new Node(name, outStream);
		newNode.setPubKey(pubKey);
		if (type == 0)
			clientNames.put(name, newNode);
		else
			helperNames.put(name, newNode);
	}
	
	public void removeAnon(String name)
	{ //Dangerous method, only used by NewThreadAnon to remove anonymous
		//that is not in chat
		if (this.checkExist(name))
			clientNames.remove(name);
	}
	
	public void takeOut(String name)
	{ //Remove a user from hashtable
		//System.out.println("Remove: " + name);
		Node tempNode;
		tempNode = clientNames.remove(name);
		if (tempNode != null)
			this.breakChat(tempNode);
		else 
			this.breakChat(helperNames.remove(name));

	}
	
	public boolean checkExist(String name)
	{ //Checks if username is already in the hastable, name must be unique
		return (!(clientNames.get(name) == null) || !(helperNames.get(name) == null));
	}
	
	public DataOutputStream getOutStream(String name)
	{
		if (checkExist(name))
		{
			return (helperNames.get(name) == null) ? clientNames.get(name).getOutput() : helperNames.get(name).getOutput();
		}
		else
			return null;
	}
	
	public PublicKey getPublicKey(String name)
	{
		if (checkExist(name))
		{
			return (helperNames.get(name) == null) ? clientNames.get(name).getPubKey() : helperNames.get(name).getPubKey();
		}
		else
			return null;
	}
	
	public boolean checkChat(Node node) /** May not be necessary, and just use chatCompanion **/
	{ //Given a node (Anon), return boolean if it is in chat
		if (node == null) return false;
		return chats.containsKey(node);
	}
	
	public Node chatCompanion(Node node)
	{ //Given Anon name, returns associated HDU (used by Anonymous users)
		return (checkChat(node)) ? chats.get(node) : null;
	}
	
	public void setChat(String name, Node node)
	{ //Name of chat starter (Anon) and node (HDU) watched to chat with
		Node chatStarter = clientNames.get(name); //Ensure not null?
		//chatStarter.setNext(node);
		node.setCount(node.getCount() + 1);
		chats.put(chatStarter, node);
		this.sendEncrypt(node.getOutput(), 
				"Server:User '"+ name + "' has been assigned to you.", node.getPubKey());
	}
	
	public void breakChat(Node node)
	{ 
		Node tempNode;
		
		tempNode = this.chatCompanion(node);
		if (tempNode != null)
		{ //Person that broke chat is Anonymous
			this.sendEncrypt(tempNode.getOutput(), 
					"Server:"+node.getName()+" has left the chat.", tempNode.getPubKey());
			tempNode.setCount(tempNode.getCount() - 1);
			chats.remove(node);
		}
		else
		{ //HDU left chat, inform all Anonymous users
			List<Node> holder = new ArrayList<Node>();
			Set<Node> set = chats.keySet();
			Iterator<Node> itr = set.iterator();
			while (itr.hasNext())
			{
				tempNode = itr.next();
				if (chats.get(tempNode).getName().equals(node.getName()))
				{
					this.sendEncrypt(tempNode.getOutput(), 
							"Server:"+node.getName()+" has left the chat.", tempNode.getPubKey());
					holder.add(tempNode);
				}
			}
			for (Node removeNode : holder)
			{
				chats.remove(removeNode);
			}
		}
	}
	
	public Node returnNode(String name)
	{ //Gives direct access to the Node holding connected user information
		//Node tempNode = clientNames.get(name);
		return clientNames.get(name);
	}
	
	public Node returnHDU(String name)
	{ //Returns node belonging to a authenticated HDU
		return helperNames.get(name);
	}
	
	public String list(DataOutputStream outStream, Integer num)
	{ //num is used to signify, 0 = server request, 1 = user request 
		try 
		{
			DataOutputStream tempStream;
			String str = "";
			
			/*if (num != 0) //If requested by a user and not server
			{
				outStream.writeBytes("List of currently connected " +
					"users requested..." + '\n');
			}*/
			Set<String> set = helperNames.keySet();
			Iterator<String> itr = set.iterator();
			String listOfNames = "";
			while (itr.hasNext())
			{
				str = itr.next();
				if (num != 0)
				{ //User request, just print the names on their terminal
					//outStream.writeBytes("Server:List:"+ str + '\n');
					listOfNames = listOfNames+":"+str;
				}
				else
				{ //Server request to shutdown, inform users and have them
					//gracefully shutdown as much as possible
					tempStream = this.getOutStream(str);
					this.sendEncrypt(tempStream, "" + '\n' + "Server:Server is shutting down, " +
						"program will now exit" + '\n' + "Server:Good Bye", this.getPublicKey(str));
					//this.takeOut(str);
				}
			}
			if (num == 1)
			{
				//outStream.writeBytes(this.encrypt("Server:List:" + listOfNames, pubKey) + '\n');
			}
			else if (num == 2)
			{
				return listOfNames.substring(1);
			}
		}
		catch (Exception e)
		{ 
			//System.out.println(e); //Debugging purposes only
		}
		return null;
	}
	
	public Node seekChat()
	{ //Finds a HDU with the lowest chat count value 
		Collection<Node> set = helperNames.values();
		Iterator<Node> itr = set.iterator();
		int count = 0;
		boolean first = true;
		Node send = null;
		Node temp = new Node();
		
		while (itr.hasNext())
		{
			temp = itr.next();
			if (first)
			{
				send = temp;
				count = temp.getCount();
				first = false;
			}
			send = (count < temp.getCount()) ? send : temp;
			count = (count < temp.getCount()) ? count : temp.getCount();
		}
		return send;
	}
	
	public void transfer(Node Anon, Node HDU)
	{ //Transfers an Anonymous user to a different HDU
		Node previousHDU = this.chatCompanion(Anon);
		this.chatCompanion(Anon).setCount(this.chatCompanion(Anon).count - 1);
		chats.remove(Anon);
		chats.put(Anon, HDU);
		HDU.setCount(HDU.count + 1);
		this.sendEncrypt(Anon.getOutput(), 
				"Server:You have been transferred to: '" + HDU.getName() + "'.", Anon.getPubKey());
		this.sendEncrypt(HDU.getOutput(),"Server:User '"+ 
				Anon.getName() + "' has been transferred to you.", HDU.getPubKey());
		this.sendEncrypt(previousHDU.getOutput(), 
				"Server:"+Anon.getName()+ " has been successfully transferred.", previousHDU.getPubKey());
	}
	
	public void updateHDUList()
	{
		String listOfNames = this.list(null, 2);
		DataOutputStream outStream;
		String str = "";
		
		Set<String> set = helperNames.keySet();
		Iterator<String> itr = set.iterator();
		
		while (itr.hasNext())
		{
			str = itr.next();
			outStream = this.getOutStream(str);
			this.sendEncrypt(outStream, "Server:List:" + listOfNames, this.getPublicKey(str));
		}
	}
	
	public void kill()
	{
		DataOutputStream tempStream;
		String str = "";
		try 
		{
			Set<String> set = clientNames.keySet();
			Iterator<String> itr = set.iterator();
			while (itr.hasNext())
			{
				str = itr.next();
				tempStream = this.getOutStream(str);
				this.sendEncrypt(tempStream, "" + '\n' + "Server:Server is shutting down, " +
						"program will now exit" + '\n' + "Server:Good Bye", this.getPublicKey(str));
			}
			
			set = helperNames.keySet();
			itr = set.iterator();
			while (itr.hasNext())
			{
				str = itr.next();
				tempStream = this.getOutStream(str);
				this.sendEncrypt(tempStream, "" + '\n' + "Server:Server is shutting down, " +
						"program will now exit" + '\n' + "Server:Good Bye", this.getPublicKey(str));
			}
			System.exit(0);
		}
		catch (Exception e) {}
	}
	
	private static void usage()
	{ //In case ChatServer is called in an improper format
		System.out.println("Proper usage is:");
		System.out.println("java -jar <Path to jar> <port number>");
		System.exit(0);
	}

	public void removeUnknown(String ipAddress, Socket connectionSocket)
	{
		if (unknownNames.containsKey(ipAddress))
		{
			if (unknownNames.get(ipAddress).getUnknown() == 1)
			{
				unknownNames.remove(ipAddress);
			}
			else
			{
				int temp = unknownNames.get(ipAddress).getConnectList().indexOf(connectionSocket);
				if (temp != -1)
				{
					unknownNames.get(ipAddress).getConnectList().remove(temp);
					unknownNames.get(ipAddress).setUnknown(unknownNames.get(ipAddress).getUnknown() - 1);
				}
			}
		}
	}

	public boolean validate(String name, String passwd)
	{
		String line = "";
		try 
		{
			FileReader fileReader = new FileReader(authFile);
			BufferedReader bufferedReader = new BufferedReader(fileReader);

			while((line = bufferedReader.readLine()) != null) 
			{
				String[] temp = line.split(":", 2);
				if (temp[0].equals(name) && temp[1].equals(passwd))
				{
					bufferedReader.close();
					return true;
				}
			}	
			bufferedReader.close();
		}
		catch (FileNotFoundException e)
		{
			System.out.println("Authentication file missing");
			
		} catch (Exception e)
		{
			System.out.println("Error reading authorization file");
		}
		return false;
	}

	public void incrementBlock(String ipAddress)
	{
		//System.out.println("incrementBlock");
		if (!blockList.containsKey(ipAddress))
		{
			Node n = new Node();
			n.setAttempts(1);
			n.setIpAddress(ipAddress);
			blockList.put(ipAddress, n);
		}
		else 
		{ //Allow it to flow past the config file for logging purposes later
			blockList.get(ipAddress).setAttempts(blockList.get(ipAddress).getAttempts() + 1);
			if (blockList.get(ipAddress).getAttempts() > maxConnectionAttempts)
			{
				blockList.get(ipAddress).setLockTimer(System.currentTimeMillis());
			}
		}
	}

	public boolean blocked(String ipAddress)
	{
		if (blockList.get(ipAddress) != null)
			return ((System.currentTimeMillis() - blockList.get(ipAddress).getLockTimer())) < lockOutTime * 1000 ?
				true : false;
		else
			return false;
	}
	
	public void sendEncrypt (DataOutputStream out, String text, PublicKey pubKey2)
	{
		try
		{
			byte[] temp = ch.encrypt(text, pubKey2);
			out.writeBytes(""+temp.length+'\n');
			out.write(temp);
			out.flush();
		}
		catch (IOException e)
		{} 
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	public String decrypt(String temp)
	{ //squelch all exceptions for now till logging
		try
		{
			return new String(ch.decrypt(temp.getBytes()));
		} catch (Exception e)
		{
			e.printStackTrace();
		}
		return null;
	}

	public PublicKey getServerPublic()
	{
		return this.pubKey;
	}
	
	public String readIncoming(BufferedReader inFromClient,
			Socket connectionSocket)
	{
		int len;
		try
		{
			len = Integer.parseInt(inFromClient.readLine());
			byte[] tempBytes = new byte[len];
			for (int i = 0; i < len; i++)
			{
				tempBytes[i] = (byte) connectionSocket.getInputStream().read();
			}
			return ch.decrypt(tempBytes);
		} catch (NumberFormatException e)
		{
			e.printStackTrace();
		} catch (IOException e)
		{
			e.printStackTrace();
		} catch (Exception e)
		{
			e.printStackTrace();
		}
		System.out.println("readIncoming (Server) is returning a NULL value!");
		return null;
	}

	public PublicKey getClientKey(BufferedReader inFromClient,
			Socket connectionSocket)
	{
		KeyFactory keyFactory;
		try
		{
			int len;
			len = Integer.parseInt(inFromClient.readLine());
			byte[] tempBytes = new byte[len];
			for (int i = 0; i < len; i++)
			{
				tempBytes[i] = (byte) connectionSocket.getInputStream().read();
			}
				
			keyFactory = KeyFactory.getInstance("RSA");
			KeySpec keySpec = new X509EncodedKeySpec(tempBytes);
		    return keyFactory.generatePublic(keySpec);
		} catch (NoSuchAlgorithmException e)
		{
			//e.printStackTrace();
		} catch (InvalidKeySpecException e)
		{
			//e.printStackTrace();
		} catch (IOException e)
		{
			e.printStackTrace();
		}
		System.out.println("getClientKey (Server) is returning a NULL value!");
		return null;
	}

	public void sendPublicKey(DataOutputStream outToClient) throws IOException
	{
		byte[] publicKeyBytes = pubKey.getEncoded();
		outToClient.writeBytes(publicKeyBytes.length + "" + '\n');
		outToClient.write(publicKeyBytes);
		outToClient.flush();
	}
}