package serverSide;

import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.InputStreamReader;
import java.net.Socket;
import java.nio.CharBuffer;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.X509EncodedKeySpec;

/* New class to handle multiple connections to the same IP, so that 
the Main() will still be listening on the set port */

class NewThread implements Runnable 
{
	private Socket connectionSocket;
	private ChatServer server;
	private DataOutputStream outToClient;
	private BufferedReader inFromClient;
	private String outSentence = "";
	private String clientSentence = "";
	boolean logged = false; //Used as a constant reminder user is logged in or not
	private String name = "";
	private PublicKey pubKey;	
	
	NewThread(ChatServer server, Socket connectionSocket, BufferedReader inFromClient) 
	{
		this.server = server; //Needed as we will be calling upon methods which
		//will require us to know who spawned this thread
		this.connectionSocket = connectionSocket;//pass connection socket
		this.inFromClient = inFromClient;
	}
	
	public void run()
	{
		try 
		{
			outToClient = new DataOutputStream(connectionSocket.getOutputStream());
			//Get public key from Client-side
			pubKey = server.getClientKey(inFromClient, connectionSocket);
			server.sendPublicKey(outToClient);
			
			registration();
			
			/* dealing with List/Chat/./Logout */
			
			//keep reading input from user until \LOGOUT
			//clientSentence = inFromClient.readLine();
			clientSentence = server.readIncoming(inFromClient, connectionSocket);
			boolean loop = true;
			
			while (loop)
			{
				/*if (clientSentence.equals("\\LIST"))
				{ //Temporary means to see online, later built into GUI
					server.list(outToClient, 1);
				}
				else if (clientSentence.equals("\\kill server"))
				{ // Temporary means to kill all sessions 
					server.kill();
					loop = false;
				}
				else */if (clientSentence.equals("\\LOGOUT"))
				{ // Check if in chat first if building this function 
					logOut();
					loop = false;
				}
				if (loop) //Only need to read again if loop wasn't broken
				{
					sendChat(clientSentence);
					//clientSentence = inFromClient.readLine();
					clientSentence = server.readIncoming(inFromClient, connectionSocket);
				}
			}
		}
		catch (Exception e)//Catch exception if any
		{
			//System.out.println("IOException on socket listen: " + ioe);
			//ioe.printStackTrace();
			//Debugging purposes only
		}
	}
	
	//Upon entering, user cannot leave until valid username is given
	private void registration()
	{ 
		try
		{
			//name = inFromClient.readLine();
			name = server.readIncoming(inFromClient, connectionSocket);
			String passwd = server.readIncoming(inFromClient, connectionSocket);
			//System.out.println("Server got name: " + name);
			//System.out.println("Server got pass: " + passwd);
			String ipAddress = connectionSocket.getRemoteSocketAddress().toString();
			ipAddress = ipAddress.split(":", 2)[0].toString().substring(1);
			
			//store name + out/in streams
			if (!(server.checkExist(name)) && server.validate(name, passwd) && !server.blocked(ipAddress))
			{ //If information is good, add to hashtable and tell client
				server.putInTable(name, outToClient, inFromClient, 1, pubKey);
				successful();
			}
			else if (!server.blocked(ipAddress))
			{ //information is already in hashtable or incorrect, tell user to try again
				server.incrementBlock(ipAddress);
				failedLogin();
			}
			else if (server.blocked(ipAddress))
			{
				server.incrementBlock(ipAddress);
				blockedLogin();
			}
		}
		catch (Exception e) {}
	}
	
	private void successful()
	{
		try
		{ //Inform user of server commands after successfully registering
			//outToClient.writeBytes(server.encrypt("Server:SUCCESSFUL REGISTRATION", pubKey) + '\n');
			server.sendEncrypt(outToClient, "Server:SUCCESSFUL REGISTRATION", pubKey);
			server.updateHDUList();
			logged = true;
		}
		catch (Exception e) {e.printStackTrace();}
	}
	
	private void failedLogin()
	{
		try
		{ //Inform user the username or password is incorrect
			//outToClient.writeBytes(server.encrypt("Server:Incorrect username or password", pubKey) + '\n');
			server.sendEncrypt(outToClient, "Server:Incorrect username or password", pubKey);
			registration();
		}
		catch (Exception e) {}
	}
	
	private void blockedLogin()
	{
		try
		{ //Inform user too many login attempts, account is locked for x minutes
			//outToClient.writeBytes(server.encrypt("Server:Too many login attempts, account has been locked for " + 
					//server.lockOutTime + " minutes.", pubKey)+ '\n');
			server.sendEncrypt(outToClient, "Server:Too many login attempts, account has been locked for " + 
					server.lockOutTime + " minutes.", pubKey);
			registration();
		}
		catch (Exception e) {}
	}
	
	private void sendChat(String text)
	{ //Deals with sending chat 
		Node sendingNode;
		DataOutputStream tempOut;
		
		try
		{
			//System.out.println("HDU Thread recieved = " + text);
			String temp[] = text.split(":", 2);
			sendingNode = server.returnNode(temp[0]);
			
			/** Could be written as (&& || &&) as msg sending process does not change **/
			if ((server.checkChat(sendingNode)) && (server.chatCompanion(sendingNode).getName().equals(this.name)))
			{ //Chat is towards an Anonymous user
				String tempStr = name + ": " + temp[1];
				tempOut = sendingNode.getOutput();
				//tempOut.writeBytes(server.encrypt(tempStr, pubKey) + '\n');
				server.sendEncrypt(tempOut, tempStr, sendingNode.getPubKey());
			}
			else if ((sendingNode == null) && (server.returnHDU(temp[0]) != null))
			{ //Chat not directed at Anonymous, but at HDU
				sendingNode = server.returnHDU(temp[0]);
				String tempStr = name + ": " + temp[1];
				tempOut = sendingNode.getOutput();
				//tempOut.writeBytes(server.encrypt("ServerCmd:"+name, pubKey)+ '\n');
				server.sendEncrypt(tempOut, "ServerCmd:"+name, sendingNode.getPubKey());
				//tempOut.writeBytes(server.encrypt(tempStr, pubKey) + '\n');
				server.sendEncrypt(tempOut, tempStr, sendingNode.getPubKey());
			}
			else if ((server.returnHDU(temp[0].split("-", 2)[0]) != null) && 
					(temp[0].split("-", 2)[1].equals("transfer")) &&
					(server.checkChat(server.returnNode(temp[1]))) &&
					(server.chatCompanion(server.returnNode(temp[1])).getName().equals(this.name)) &&
					!(server.returnHDU(temp[0].split("-", 2)[0]).getName().equals(this.name)))
			{ //Chat comes in the form "HDU01-transfer:Anonymous00"
				server.transfer(server.returnNode(temp[1]), server.returnHDU(temp[0].split("-", 2)[0]));
			}
		}
		catch (Exception e) {System.out.println(e);}
	}
	
	private void logOut()
	{
		try
		{ //Tells client to begin shut down procedures and removes username
			//from hashtable
			//outToClient.writeBytes(server.encrypt("Server:Good Bye", pubKey) + '\n');
			server.sendEncrypt(outToClient, "Server:Good Bye", pubKey);
		}
		catch (Exception e) {}
		if (server.checkExist(name))
		{
			server.takeOut(name);
			server.updateHDUList();
		}
	}
}