package serverSide;

import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.Socket;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.X509EncodedKeySpec;

/* New class to handle multiple connections to the same IP, so that 
the Main() will still be listening on the set port */

class NewThreadAnon implements Runnable 
{
	private Socket connectionSocket;
	private ChatServer server;
	private DataOutputStream outToClient;
	private BufferedReader inFromClient;
	private String outSentence = "";
	private String clientSentence = "";
	private String name = "";
	private PublicKey pubKey;
	
	NewThreadAnon(ChatServer server, Socket connectionSocket, BufferedReader inFromClient) 
	{
		this.server = server; //Needed as we will be calling upon methods which
		//will require us to know who spawned this thread
		this.connectionSocket = connectionSocket;//pass connection socket
		this.inFromClient = inFromClient;
	}
	
	public void run()
	{
		try 
		{
			outToClient = new DataOutputStream(connectionSocket.getOutputStream());
			//Get public key from Client-side
			pubKey = server.getClientKey(inFromClient, connectionSocket);
			server.sendPublicKey(outToClient);
			
			setAnonymous();
			
			Node deskNode = server.seekChat();

			if (deskNode != null)
			{ //Ensure there is an HDU to chat with
				outSentence = "Server:Welcome to the Panda Chat server lobby " + this.name + "!";
				//outToClient.writeBytes(server.encrypt(outSentence, pubKey) + '\n');
				server.sendEncrypt(outToClient, outSentence, pubKey);
				//outToClient.writeBytes(server.encrypt("Server:Help-Desk Agent " + deskNode.getName() + 
						//" has been assigned to you.", pubKey) + '\n');
				server.sendEncrypt(outToClient, "Server:Help-Desk Agent " + deskNode.getName() + 
						" has been assigned to you.", pubKey);
				server.setChat(this.name, deskNode);
				
				//clientSentence = inFromClient.readLine();
				clientSentence = server.readIncoming(inFromClient, connectionSocket);
				while (!clientSentence.equals("\\LOGOUT"))
				{
					sendChat(clientSentence);
					//clientSentence = inFromClient.readLine();
					clientSentence = server.readIncoming(inFromClient, connectionSocket);
				}
				logOut();
			}
			else
			{ //No HDU can be chatted with
				//outToClient.writeBytes(server.encrypt("Server:Unable to locate any Help-Desk users to chat with", 
						//pubKey) + '\n');
				server.sendEncrypt(outToClient, "Server:Unable to locate any Help-Desk users to chat with", pubKey);
				this.removeAnonymous();
			}
			//logOut(); //Uncommenting causes client to shut down immediately upon failed connect
		}
		catch (Exception e)//Catch exception if any
		{
			//System.out.println("IOException on socket listen: " + e);
			//e.printStackTrace();
			//Debugging purposes only
		}
	}

	private void setAnonymous()
	{
		int i = 0;
		
		while (server.checkExist("Anonymous"+i))
		{
			i++;
		}
		//System.out.println("Unused name: " + i);
		this.name = "Anonymous" + i;
		
		server.putInTable(this.name, outToClient, inFromClient, 0, pubKey);
	}
	
	private void removeAnonymous()
	{
		server.removeAnon(name);
	}
	
	private void sendChat(String text)
	{ //Deals with sending chat 
		Node sendingNode;
		DataOutputStream tempOut;
		
		try
		{
			if (server.checkChat(server.returnNode(name)))
			{
				sendingNode = server.chatCompanion(server.returnNode(name));
				if (text.split(":", 2)[0].equals(sendingNode.getName()))
				{
					String tempStr = name + ": " + text.split(":", 2)[1];
					tempOut = sendingNode.getOutput();
					//tempOut.writeBytes(server.encrypt(tempStr, pubKey) + '\n');
					server.sendEncrypt(tempOut, tempStr, 
							server.chatCompanion(server.returnNode(name)).getPubKey());
				}
			}
		}
		catch (Exception e) {System.out.println(e);}
	}
	
	private void logOut()
	{
		try
		{ //Tells client to begin shut down procedures and removes username
			//from hashtable
			//outToClient.writeBytes(server.encrypt("Server:Good Bye", pubKey) + '\n');
			server.sendEncrypt(outToClient, "Server:Good Bye", pubKey);
		}
		catch (Exception e) {}
		if (server.checkExist(name))
			server.takeOut(name);
	}
}